"""
The functions in this file return true if a chapter is found for the specific site
They take in as input the html of the webpage
"""
from __future__ import annotations
from typing import Callable


__all__ = ("domains",)


#
# Return true if chapter is found
#


def mangakakalot_tv(data: str) -> bool:
    return "next-chap:hover" in data and "logo_chapter" in data


def mangakakalot_com(data: str) -> bool:
    return (
        "nextchap.png" not in data
        and "Sorry, the page you have requested cannot be found." not in data
        and "PREV CHAPTER" in data
    )


def mangakakalots_com(data: str) -> bool:
    return (
        "500 Internal Server Error" not in data
        and "Sorry, the page you have requested cannot be found. Click" not in data
    )


def mangakakalot_cc(data: str) -> bool:
    return "nextchap.png" not in data and "is not available yet" not in data


def mangakakalot_xyz(data: str) -> bool:
    return "View: all images" in data


def mangakakalot_city(data: str) -> bool:
    return "Next Chapter:" in data and "is not available yet. We will update" not in data


def mangakakalot_online(data: str) -> bool:
    return "http://mangakakalot.online/frontend/images/nextchap.png" not in data


def manganelo_com(data: str) -> bool:
    return "404 - PAGE NOT FOUND" not in data and "This page does not exist or has been deleted." not in data


def manganelos_com(data: str) -> bool:
    return "is not available yet" not in data and "next_img" in data


def mangatx_com(data: str) -> bool:
    return "https://mangatx.com/wp-content/uploads/WP-manga/data/manga" in data and "image-0" in data


def mangakakalot_today(data: str) -> bool:
    return (
        "http://mangakakalot.today/frontend/imgs/nextchap.png" not in data
        and "is not available yet. We will update " not in data
    )


def chapmanganato_to(data: str) -> bool:
    return "404 NOT FOUND" not in data and "LOAD ALL" in data


def chapmanganato_com(data: str) -> bool:
    return "404 NOT FOUND" not in data and "LOAD ALL" in data


def mangabuddy_com(data: str) -> bool:
    return '"Next chapter"' in data and '"Previous chapter"' in data


def mangaclash_com(data: str) -> bool:
    return "cursorNext" in data and "Alternative" not in data


#
# Known domains
#


domains: dict[str, Callable[[str], bool]] = {
    "mangatx.com": mangatx_com,
    "mangaclash.com": mangaclash_com,
    "mangabuddy.com": mangabuddy_com,
    #
    "manganelo.com": manganelo_com,
    "manganelos.com": manganelos_com,
    #
    "mangakakalot.tv": mangakakalot_tv,
    "mangakakalot.cc": mangakakalot_cc,
    "mangakakalot.xyz": mangakakalot_xyz,
    "mangakakalot.com": mangakakalot_com,
    "mangakakalot.city": mangakakalot_city,
    "mangakakalot.today": mangakakalot_today,
    "mangakakalot.online": mangakakalot_online,
    #
    "mangakakalots.com": mangakakalots_com,
    "mangakakalots.net": mangakakalots_com,
    #
    "readmanganato.com": chapmanganato_com,
    "chapmanganato.com": chapmanganato_com,
    "chapmanganato.to": chapmanganato_to,
}



from __future__ import annotations
from concurrent.futures import Future, thread
from typing import Callable, Any
from functools import lru_cache
from pathlib import Path
import subprocess
import traceback
import argparse
import platform
import time
import sys

from tqdm import tqdm
import requests

from manga.utils import extract_url, lsf, split_on_num
from manga import sites


class ThreadHandler(thread.ThreadPoolExecutor):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.futures: list[Future] = []

    def add(self, fn: Callable[..., Any], *args: Any, **kwargs: Any) -> None:
        self.futures.append(super().submit(fn, *args, **kwargs))

    def kill(self):
        self.shutdown(wait=False)
        for i in self.futures:
            i.cancel()


#
# Failure Classes
#


class Failed:
    reason: str = ""  # Set by subclasses
    _unique_reason: bool = False  # Overridden by subclasses if desired

    def __init__(self, url: str):
        self.url: str = url

    def __str__(self):
        if self._unique_reason:
            return f"{self.reason}: {self.url}"
        return self.url

    def __repr__(self):
        ret = f"<{self.__class__.__name__}: url='{self.url}'"
        if self._unique_reason:
            ret += f", reason='{self.reason}'"
        return ret + ">"

    @classmethod
    def kind(cls):
        if cls._unique_reason:
            return cls.__name__
        return f"{cls.__name__}: {cls.reason}"


# Should not happen


class NoOpen(Failed):
    pass


class Unknown(NoOpen):
    reason = "This website is for an unknown / unsupported domain."


class BadRequest(NoOpen):
    _unique_reason = True

    def __init__(self, url: str, reason):
        super().__init__(url)
        self.reason = f"A request failed: {reason}"


class MalformedURL(NoOpen):
    pass


class NotInt(MalformedURL):
    reason = "The URL's chapter is not an integer."


class HasVol(MalformedURL):
    reason = "The URL contains 'vol'; this is a bad sign"


class Pattern(MalformedURL):
    reason = "The URL contains a pattern that is dangerous"


# Non-Malformed URL Failure Classes


class ToOpen(Failed):
    pass


class Tiny(ToOpen):
    reason = "URL failed by default, it is too small"


class Exists(ToOpen):
    reason = "The URL is valid, but this site seems to be missing other chapters"


class Missing(ToOpen):
    reason = "Previous and future chapters exist, this one does not."


class Broken(ToOpen):
    reason = "This website does not seem to have any chapters of this manga."


######################################################################
#                           Main Functions                           #
######################################################################


@lru_cache(maxsize=None)
def _test_site(url: str):
    return sites.test(url)


def _test(left: str, right: str, x: float) -> bool:
    """
    Test if chapter x is found at a URL constructed from left, right, and x
    """
    if int(x) == x:
        return _test_site(f"{left}{int(x)}{right}")
    else:
        if _test_site(f"{left}{x}{right}"):
            return True
        return _test_site(f"{left}{str(x).replace('.', '-')}{right}")


# pylint: disable=too-many-return-statements,too-many-branches
def _evaluate(url: str, delay: int) -> Failed | None:
    """
    If url is broken, return a failure class for it
    """
    if "vol" in url.lower():
        return HasVol(url)
    left, n, right = split_on_num(url)
    if n != int(n):
        return NotInt(url)
    elif n <= 5:
        return Tiny(url)
    elif "mangabuddy" in left and ("/mbx" in left or any(i.isalpha() for i in right)):
        return Pattern(url)
    test = lambda x: _test(left, right, x)
    try:
        if test(n) and not test(n - 1) and not test(5):
            return Exists(url)
        time.sleep(0.25)  # No DOS-ing
        if not test(n):
            for i in (0.1, 1, 1.1, 2, 2.1, 5, 10, 20):
                if test(n + i):
                    return Missing(url)
        time.sleep(0.5)  # No DOS-ing
        if not any(test(i) for i in (n, n - 1, 5, n + 0.1, n + 1, n + 1.1)):
            return Broken(url)
        time.sleep(0.25)  # No DOS-ing
        return None
    except sites.UnknownDomain:
        return Unknown(url)
    except requests.exceptions.RequestException as e:
        return BadRequest(url, e)
    finally:
        time.sleep(delay)  # No DOS-ing


def evaluate(url: str, delay: int, ret: list, pbar) -> None:
    """
    Determine if url is broken
    Store return value in ret because we the return value will be ignored
    """
    try:
        rv: Failed | None = _evaluate(url, delay)
        if rv is not None:
            ret.append(rv)
    except:
        traceback.print_exc()
        raise
    finally:
        pbar.update()


def print_each(prefix: str, lst: list[Failed]):
    """
    Print every item in lst
    """
    if lst:
        lst = sorted(lst, key=lambda x: x.url)
        print(f"{prefix}\n\t" + "\n\t".join(str(i) for i in lst))


def open_each(opener: str, prefix: str, lst: list[Failed]) -> None:
    """
    Open every item in lst
    """
    if lst:
        print(f"{prefix}\n\t" + "\n\t".join(i.url for i in lst))
        for i in tqdm(lst, leave=False):
            subprocess.check_call([opener, i.url], stdout=subprocess.DEVNULL)
        print("")


def test_sites(directory: Path, opener: str, prompt: bool, n_workers: int, delay: int) -> bool:
    """
    Test each file in directory, print the results open them as needed
    """
    print("Checking arguments...")
    directory = directory.resolve()
    assert delay >= 0, "Delay may not be negative"
    assert n_workers > 0, "n_workers must be positive"
    assert directory.exists(), f"{directory} does not exist"
    assert directory.is_dir(), f"{directory} is not a directory"
    # Determine which requests must be made
    print("Scanning files...")
    urls: set[str] = {extract_url(i) for i in lsf(directory)}
    # Determine what to open
    tested: list[Failed] = []
    print(f"Testing {len(urls)} urls using {n_workers} workers...")
    with tqdm(total=len(urls)) as pbar:
        with ThreadHandler(max_workers=n_workers) as executor:  # No DOS-ing
            for i in urls:
                executor.add(evaluate, i, delay, tested, pbar)
    # Results
    no_open: list[Failed] = [i for i in tested if isinstance(i, NoOpen)]
    to_open: list[Failed] = [i for i in tested if isinstance(i, ToOpen)]
    # Print no-open failures
    sub_list: Callable[[list[Failed], type], list[Failed]] = lambda lst, t: [i for i in lst if isinstance(i, t)]
    if no_open:
        print(f"{'*'*70}\n*{'Errors'.center(68)}*\n{'*'*70}\n")
        for sub in {i.__class__ for i in no_open}:
            print_each(sub.kind(), sub_list(no_open, sub))
        print(f"\n{'*'*70}\n*{'Done'.center(68)}*\n{'*'*70}\n")
    # Open
    if prompt:
        _ = input("Testing complete. Hit enter to open websites.")
    for sub in {i.__class__ for i in to_open}:
        open_each(opener, sub.kind(), sub_list(to_open, sub))
    return True


def main(prog: str, *args: str) -> bool:
    assert "Darwin" == platform.system(), "Not on Mac! Remember to change name and ext!"
    parser = argparse.ArgumentParser(prog=Path(prog).name)
    parser.add_argument("directory", type=Path, help="The directory to test")
    parser.add_argument("--n_workers", default=16, type=int, help="The number of sites to test concurrently")
    parser.add_argument("--opener", default="open", help="The default binary to open a URL with")
    parser.add_argument(
        "--prompt", action="store_true", help="Do not auto open sites when complete, prompt user instead"
    )
    parser.add_argument(
        "--delay",
        default=1,
        type=int,
        help="The number of seconds each thread should wait between testing sites (to avoid DOSing)",
    )
    return test_sites(**vars(parser.parse_args(args)))


def cli() -> None:
    sys.exit(0 if main(*sys.argv) else -1)



from __future__ import annotations
from pathlib import Path
import collections
import argparse
import platform
import time
import sys

import osascript

from manga.utils import split_on_num, extract_url, lsf


def scrub(url: str) -> str:
    """
    Scrub a URL for easy comparison with others
    """
    url = "".join(url.split("//")[1:])
    if "." in url[:4]:
        url = url.split(".", 1)[1]
    if url.endswith("/"):
        url = url[:-1]
    if "?" in url:
        url = url.split("?", 1)[0]
    return url


def read_dir(directory: Path) -> dict[str, str]:
    """
    Read a directory and map the scrubbed URLs to their chapter numbers a strings
    Ignores .swp files but whines about them
    """
    assert directory.is_dir(), f"{directory} is a file"
    files: tuple[Path, ...] = lsf(directory)
    ret: dict[str, str] = {}
    for f in files:
        fname: str = f.name
        if fname.startswith(".") and f.suffix == ".swp":
            print(f"Refusing to open swap file: {f}")
            continue
        _, num, _2 = split_on_num(fname)
        ret[scrub(extract_url(f))] = str(num)
    return ret


def get_url() -> str:
    """
    Get the URL of the current tab
    """
    cmd: str = 'tell application "Google Chrome" to return URL of active tab of front window'
    code, out, err = osascript.run(cmd)
    assert code == 0, err
    return out


def unnumbered_helper(dirs: list[Path]) -> None:
    """
    Print out the number of the chapter associated
    with the manga displayed in the frontmost Chrome tab
    """
    for i in dirs:
        assert i.exists(), f"{i} does not exist"
    raw_data: list[dict[str, str]] = [read_dir(i) for i in dirs]
    data: dict[str, str] = dict(collections.ChainMap(*raw_data))
    old: str | None = None
    while True:
        time.sleep(0.05)
        url: str = scrub(get_url())
        if url == old:
            continue
        old = url
        try:
            print(data[url])
        except KeyError:
            print("Unknown")


def main(prog: str, *argv: str) -> bool:
    assert "Darwin" == platform.system(), "Not on Mac! Remember to change name and ext!"
    parser = argparse.ArgumentParser(prog=Path(prog).name)
    parser.add_argument("dirs", type=Path, nargs="+", help="The unnumbered directories")
    try:
        unnumbered_helper(**vars(parser.parse_args(argv)))
    except KeyboardInterrupt:
        pass
    return True


def cli() -> None:
    sys.exit(0 if main(*sys.argv) else -1)



from __future__ import annotations
from dataclasses import dataclass, astuple, field
from concurrent.futures import Future, thread
from typing import Callable, Any
from pathlib import Path
import subprocess
import argparse
import platform
import signal
import time
import sys
import os

import requests
import tqdm

from manga.utils import extract_url, lsf
from manga import sites


mk_open_remaining_first: bool = True


@dataclass
class Tested:
    """
    The results of testing every URL
    """

    has_new: set[str] = field(default_factory=set)
    ignore: set[str] = field(default_factory=set)
    skip: set[str] = field(default_factory=set)
    failed: set[str] = field(default_factory=set)
    unknown: set[str] = field(default_factory=set)


######################################################################
#                         Threading Handlers                         #
######################################################################


class ThreadHandler(thread.ThreadPoolExecutor):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.futures: list[Future] = []

    def add(self, fn, *args: Any, **kwargs: Any) -> None:
        self.futures.append(super().submit(fn, *args, **kwargs))

    def kill(self):
        self.shutdown(wait=False)
        for i in self.futures:
            i.cancel()


def mk_open_remaining(executor, urls, tested) -> Callable[[int, Any], None]:
    """
    Return a signal handler to open the remaining tested URLs
    """

    def handler(_: int, _2: Any) -> None:
        global mk_open_remaining_first  # pylint: disable=global-statement
        if not mk_open_remaining_first:
            os._exit(1)  # pylint: disable=protected-access
        mk_open_remaining_first = False
        print("Terminating executor...")
        executor.kill()
        handle_results(urls, tested)
        os._exit(0)  # pylint: disable=protected-access

    return handler


######################################################################
#                           Main Functions                           #
######################################################################


def evaluate(url: str, tested: Tested, pbar: tqdm.std.tqdm) -> None:
    """
    Determine if url has a new chapter or not
    Store the result in tested and update pbar
    """
    try:
        (tested.has_new if sites.test(url) else tested.ignore).add(url)
    except sites.UnknownDomain:
        tested.unknown.add(url)
    except requests.exceptions.RequestException:
        tested.failed.add(url)
    finally:
        pbar.update()


def handle_results(urls: set[str], tested: Tested) -> None:
    """
    Print out the test results and open each of the given URLs that should not be ignored
    """
    if len(tested.unknown) > 0:
        print("The following domains were not known:")
        print("\t" + "\n\t".join(sorted(tested.unknown)))
    if len(tested.failed) > 0:
        print("The following domains could not be opened:")
        print("\t" + "\n\t".join(sorted(tested.failed)))
    if len(tested.skip) > 0:
        print("The following domains were skipped:")
        print("\t" + "\n\t".join(sorted(tested.skip)))
    all_tested = set().union(*astuple(tested))
    if len(all_tested) != len(urls):
        print("The following domains were not tested:")
        print("\t" + "\n\t".join(sorted(urls - all_tested)))
        print("Assuming all remaining URLs must be opened...")
    print("Opening manga...")
    for url in tqdm.tqdm(urls - tested.ignore - tested.skip):
        subprocess.check_call(["open", url], stdout=subprocess.DEVNULL)
        time.sleep(0.2)  # Rate limit


def open_new(directory: Path, skip: set[str] | list[str]) -> bool:
    """
    Open each file in directory that has a new chapter ready
    """
    if isinstance(skip, list):
        return open_new(directory, set(skip))
    print("Checking arguments...")
    directory = directory.resolve()
    assert directory.exists(), f"{directory} does not exist"
    assert directory.is_dir(), f"{directory} is not a directory"
    # Determine which requests must be made
    print("Scanning files...")
    urls: set[str] = {extract_url(i) for i in lsf(directory)}
    results = Tested()
    # Determine what to open
    print(f"Making at most {len(urls)} requests...")
    original_sigint_handler: Any = signal.getsignal(signal.SIGINT)
    with tqdm.tqdm(total=len(urls)) as pbar:
        with ThreadHandler(max_workers=32) as executor:  # No DOS-ing
            signal.signal(signal.SIGINT, mk_open_remaining(executor, urls, results))
            for i in urls:
                if sites.get_domain(i) in skip:
                    results.skip.add(i)
                    pbar.update()
                else:
                    executor.add(evaluate, i, results, pbar)
    signal.signal(signal.SIGINT, original_sigint_handler)
    # Open links
    handle_results(urls, results)
    return True


def main(prog: str, *args: str) -> bool:
    assert "Darwin" == platform.system(), "Not on Mac! Remember to change name and ext!"
    parser = argparse.ArgumentParser(prog=Path(prog).name)
    parser.add_argument("--skip", type=str, nargs="+", default=[], help="Domains to skip")
    parser.add_argument("directory", type=Path, help="The directory to open new items from")
    return open_new(**vars(parser.parse_args(args)))


def cli() -> None:
    sys.exit(0 if main(*sys.argv) else -1)
